from .bag_cross_validate import (cross_validate_bag, 
    bags_2_si, 
    bags_2_si_generator, 
    BagScorer,
    _fit_and_score)


__author__ = 'John Vorsten'
__email__ = 'vorstenjohn@gmail.com'
__version__ = '0.0.1'
